﻿using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using FarseerPhysics.Dynamics;
using FarseerPhysics.Common;
using FarseerPhysics.Common.Decomposition;
using FarseerPhysics.Factories;

namespace XFORGE.StillCharging
{
    public abstract class Entity : Body
    {
        private static Queue<Entity> spawnQueue = new Queue<Entity>();

        public static Queue<Entity> SpawnQueue
        {
            get
            {
                return spawnQueue;
            }
            private set
            { }
        }

        private static Queue<Entity> despawnQueue = new Queue<Entity>();

        public static Queue<Entity> DespawnQueue
        {
            get
            {
                return despawnQueue;
            }
            private set
            { }
        }

        private static World world;

        public static World World
        {
            get
            {
                return world;
            }
            set
            {
                world = value;
            }
        }

        private static Dictionary<string, Texture2D> textures;

        public static Dictionary<string, Texture2D> Textures
        {
            get
            {
                return textures;
            }
            set
            {
                textures = value;
            }
        }

        protected Texture2D texture;

        protected float opacity;

        protected Vector2 origin;

        /// <summary>
        /// Generates a compound polygon fixture from the entities texture and attaches
        /// it to the body.
        /// </summary>
        /// <param name="density">The density for the fixture</param>
        protected void AttachTextureFixture(float density)
        {
            // Create an array to hold the data from the texture
            uint[] data = new uint[(texture.Width) * (texture.Height)];

            // Collect data from bitmap
            texture.GetData(data);

            // Create Polygon from Bitmap
            Vertices verts = PolygonTools.CreatePolygon(data, (texture.Width), false);

            // Make sure that the origin of the texture is the centroid (real center of geometry)
            Vector2 scale = new Vector2(ConvertUnits.ToSimUnits(1), ConvertUnits.ToSimUnits(1));
            verts.Scale(ref scale);

            // Make sure that the origin of the texture is the centroid (real center of geometry)
            var polygonOrigin = verts.GetCentroid();

            // Translate the polygon so that it aligns properly with centroid.
            Vector2 vertsTranslate = -polygonOrigin;
            verts.Translate(ref vertsTranslate);

            // Decompose polygon into smaller chuncks that Farseer can process better
            List<Vertices> list;
            list = BayazitDecomposer.ConvexPartition(verts);

            FixtureFactory.AttachCompoundPolygon(list, density, this);
        }

        public Entity(Texture2D texture, Vector2 position, float rotation)
            : base(world)
        {
            this.texture = texture;
            this.origin = new Vector2(texture.Width / 2, texture.Height / 2);
            this.opacity = 1.0f;

            Position = position;
            Rotation = rotation;

            BodyType = BodyType.Dynamic;
        }

        public Entity(Texture2D texture, Vector2 position)
            : this(texture, position, 0.0f)
        { }

        public virtual void Update(GameTime gameTime)
        { }

        public virtual void Draw(SpriteBatch spriteBatch)
        {
            spriteBatch.Draw(
                texture,
                new Rectangle(
                    (int)ConvertUnits.ToDisplayUnits(Position.X),
                    (int)ConvertUnits.ToDisplayUnits(Position.Y), 
                    texture.Width,
                    texture.Height
                    ),
                null,
                Color.White * opacity,
                Rotation,
                origin,
                SpriteEffects.None,
                0.0f
                );
        }
    }
}
