﻿#region Using Statements
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;
using Microsoft.Xna.Framework.GamerServices;
using FarseerPhysics.Dynamics;
using FarseerPhysics.DebugViews;
using FarseerPhysics;
using System.IO;
#endregion

namespace XFORGE.StillCharging
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class SCGame : Game
    {
        private GraphicsDeviceManager graphics;
        private SpriteBatch spriteBatch;

        private Dictionary<string, Texture2D> textures;

        //private Texture2D playerShipTexture;
        //private Texture2D turretSmallTexture;
        //private Texture2D laserSmallBoltTexture;
        //private Texture2D asteroidNormalTexture;
        //private Texture2D starTinyTexture;
        //private Texture2D starMediumTexture;
        //private Texture2D pixelTexture;

        private SpriteFont fontStatus;

        private GameState gameState;

        private KeyboardState lastKeyboardState;
        private MouseState lastMouseState;

        private World world;
#if DEBUG
        private DebugViewXNA debugView;
#endif

        private List<Entity> entities;

        private Starfield starfield;
        private PlayerShip playerShip;

        private float verticalOffset;
        private float nextStarfieldPopulate;
        private float nextAsteroidSpawn;

        private float gameOverDelay;

        public SCGame()
            : base()
        {
            Window.Title = "Still Charging!";

            graphics = new GraphicsDeviceManager(this);

            graphics.PreferredBackBufferWidth = 1280;
            graphics.PreferredBackBufferHeight = 720;

            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            gameState = GameState.Initializing;

            textures = new Dictionary<string, Texture2D>();
            entities = new List<Entity>();
            world = new World(Vector2.Zero);
            Entity.World = world;

            verticalOffset = 0.0f;
            nextStarfieldPopulate = 0.0f;
            nextAsteroidSpawn = 0.0f;
            gameOverDelay = 0.0f;

#if DEBUG
            debugView = new DebugViewXNA(world);
            debugView.AppendFlags(DebugViewFlags.Shape);
            //debugView.AppendFlags(DebugViewFlags.PolygonPoints);
            debugView.AppendFlags(DebugViewFlags.ContactPoints);
            debugView.AppendFlags(DebugViewFlags.PerformanceGraph);
            //debugView.AppendFlags(DebugViewFlags.AABB);
            debugView.AppendFlags(DebugViewFlags.Pair);
            debugView.AppendFlags(DebugViewFlags.DebugPanel);
            debugView.AppendFlags(DebugViewFlags.CenterOfMass);
            debugView.AppendFlags(DebugViewFlags.Controllers);
#endif

            base.Initialize();

            gameState = GameState.PreGame;
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            var contentFileNames = Directory.EnumerateFiles(
                Content.RootDirectory,
                "*.png",
                SearchOption.TopDirectoryOnly
                );

            foreach (var fileName in contentFileNames)
            {
                var strippedName = Path.GetFileNameWithoutExtension(fileName);

                textures.Add(
                    strippedName,
                    Content.Load<Texture2D>(strippedName)
                    );
            }

            //playerShipTexture = Content.Load<Texture2D>("player_ship");
            //turretSmallTexture = Content.Load<Texture2D>("turret_small");
            //laserSmallBoltTexture = Content.Load <Texture2D>("laser_small_bolt");
            //asteroidNormalTexture = Content.Load<Texture2D>("asteroid_normal");
            //starTinyTexture = Content.Load<Texture2D>("star_tiny");
            //starMediumTexture = Content.Load<Texture2D>("star_medium");

            var pixelTexture = new Texture2D(GraphicsDevice, 1, 1);
            pixelTexture.SetData(new[] { Color.White });
            textures.Add("pixel", pixelTexture);

            Entity.Textures = textures;

            fontStatus = Content.Load<SpriteFont>("font_status");

#if DEBUG
            debugView.LoadContent(GraphicsDevice, Content);
#endif

            starfield = new Starfield(GraphicsDevice.Viewport, textures);
            starfield.Populate(new Rectangle(
                GraphicsDevice.Viewport.X,
                GraphicsDevice.Viewport.Y,
                GraphicsDevice.Viewport.Width,
                GraphicsDevice.Viewport.Height
                ),
                -verticalOffset
                );

            playerShip = new PlayerShip(ConvertUnits.ToSimUnits(
                new Vector2(GraphicsDevice.Viewport.Width / 2, GraphicsDevice.Viewport.Height / 2))
                );
            entities.Add(playerShip);
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Respond to user input
            HandleInput();

            switch (gameState)
            {
                case GameState.PreGame:
                    PreGameUpdate(gameTime);
                    break;
                case GameState.InGame:
                    InGameUpdate(gameTime);
                    break;
                case GameState.PostGame:
                    break;
            }

            base.Update(gameTime);
        }

        private void PreGameUpdate(GameTime gameTime)
        {
            SpawnEntities();
        }

        private void InGameUpdate(GameTime gameTime)
        {
            // Let the game world scroll down some more
            verticalOffset += 50.0f * (float)gameTime.ElapsedGameTime.TotalSeconds;

            // Add new stars and get rid of old ones
            UpdateStarfield();

            // Add some new asteroids to the scene
            SpawnAsteroids();

            // Update the physics engine
            world.Step((float)gameTime.ElapsedGameTime.TotalSeconds);

            // Update the entities
            foreach (Entity e in entities)
            {
                e.Update(gameTime);
            }

            // Spawn new entities that were created by other entities
            SpawnEntities();

            // Despawn entities that are not needed anymore
            DespawnEntities();

            if (playerShip.IsAlive)
            {
                if (playerShip.DriveEnergy == 1000.0f)
                {
                    playerShip.IsJumped = true;

                    DelayGameOver(gameTime);
                }
            }
            else
            {
                DelayGameOver(gameTime);
            }
        }

        private void DelayGameOver(GameTime gameTime)
        {
            gameOverDelay += (float)gameTime.ElapsedGameTime.TotalSeconds;

            if (gameOverDelay >= 3.0f)
                gameState = GameState.PostGame;
        }

        private void UpdateStarfield()
        {
            if (verticalOffset >= nextStarfieldPopulate)
            {
                starfield.Populate(new Rectangle(
                    0,
                    (int)-verticalOffset - GraphicsDevice.Viewport.Height,
                    GraphicsDevice.Viewport.Width,
                    GraphicsDevice.Viewport.Height
                    ),
                    -verticalOffset
                    );

                nextStarfieldPopulate += GraphicsDevice.Viewport.Height;
            }
        }

        private void HandleInput()
        {
            KeyboardState kb = Keyboard.GetState();
            MouseState ms = Mouse.GetState();

            if (kb.IsKeyDown(Keys.Escape))
                Exit();

            switch (gameState)
            {
                case GameState.PreGame:
                    if (kb.IsKeyDown(Keys.R))
                        gameState = GameState.InGame;
                    break;
                case GameState.InGame:
                    Vector2 desiredDirection = Vector2.Zero;

                    if (kb.IsKeyDown(Keys.W))
                        desiredDirection -= Vector2.UnitY;

                    if (kb.IsKeyDown(Keys.S))
                        desiredDirection += Vector2.UnitY;

                    if (kb.IsKeyDown(Keys.A))
                        desiredDirection -= Vector2.UnitX;

                    if (kb.IsKeyDown(Keys.D))
                        desiredDirection += Vector2.UnitX;

                    if (kb.IsKeyDown(Keys.R))
                        playerShip.RcsActive = true;
                    else
                        playerShip.RcsActive = false;

                    if (kb.IsKeyDown(Keys.Space))
                    {
                        playerShip.ChargingDrive = true;
                    }
                    else
                    {
                        if (playerShip.ChargingDrive)
                        {
                            playerShip.ChargingDrive = false;
                        }
                    }

                    playerShip.DesiredDirection = desiredDirection;

                    if ((lastMouseState.LeftButton == ButtonState.Released) &&
                        (ms.LeftButton == ButtonState.Pressed))
                        playerShip.Turret.Fire();

                    playerShip.Turret.Target.X = ConvertUnits.ToSimUnits(ms.X);
                    playerShip.Turret.Target.Y = ConvertUnits.ToSimUnits(ms.Y - verticalOffset);

#if DEBUG
                    if (kb.IsKeyDown(Keys.F9))
                    {
                        playerShip.ChargingDrive = true;
                        playerShip.DriveEnergy = 1000.0f;
                    }
#endif
                    break;
                case GameState.PostGame:
                    if (kb.IsKeyDown(Keys.R))
                    {
                        this.Initialize();
                    }
                    break;
            }

            lastKeyboardState = kb;
            lastMouseState = ms;
        }

        private void SpawnAsteroids()
        {
            if (verticalOffset >= nextAsteroidSpawn)
            {
                Random random = new Random();

                entities.Add(new Asteroid(ConvertUnits.ToSimUnits(
                    new Vector2(
                        random.Next(GraphicsDevice.Viewport.Width + 20) - 10,
                        -verticalOffset - 20.0f)
                    )));

                nextAsteroidSpawn += 15.0f;
            }
        }

        private void SpawnEntities()
        {
            while (Entity.SpawnQueue.Count > 0)
            {
                var e = Entity.SpawnQueue.Dequeue();

                entities.Add(e);
            }
        }

        private void DespawnEntities()
        {
            while (Entity.DespawnQueue.Count > 0)
            {
                var e = Entity.DespawnQueue.Dequeue();

                world.RemoveBody(e);
                entities.Remove(e);
            }

            var safeRectangle = new Rectangle(
                -50,
                (int)-verticalOffset - 50,
                GraphicsDevice.Viewport.Width + 100,
                GraphicsDevice.Viewport.Height + 100
                );

            for (int i = 0; i < entities.Count; i++)
            {
                var e = entities[i];

                if (!safeRectangle.Contains(
                    (int)ConvertUnits.ToDisplayUnits(e.Position.X),
                    (int)ConvertUnits.ToDisplayUnits(e.Position.Y)
                    ))
                {
                    if (e is PlayerShip)
                    {
                        ((PlayerShip)e).IsAlive = false;
                    }
                    else
                    {
                        world.RemoveBody(e);
                        entities.RemoveAt(i);
                    }
                }
            }
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            var camera = Matrix.CreateTranslation(new Vector3(0.0f, verticalOffset, 0.0f))
                //* Matrix.CreateScale(
                //    ConvertUnits.ToSimUnits(1),
                //    ConvertUnits.ToSimUnits(1),
                //    1
                //    )
                    ;

            spriteBatch.Begin(
                SpriteSortMode.Immediate,
                BlendState.NonPremultiplied,
                null,
                null,
                null,
                null,
                camera
                );

            starfield.Draw(spriteBatch, -verticalOffset);

            foreach (Entity e in entities)
            {
                e.Draw(spriteBatch);
            }

            switch (gameState)
            {
                case GameState.PreGame:
                    DrawPressStart(gameTime);

                    spriteBatch.DrawString(
                    fontStatus,
                    String.Format(
                        "Still Charging!",
                        playerShip.Turret.Energy
                        ),
                    new Vector2(GraphicsDevice.Viewport.Width / 2 - 75, 125 - verticalOffset),
                    Color.White);

                    spriteBatch.DrawString(
                        fontStatus,
                        "Movement: <W> <A> <S> <D>\nAim turret: Mouse pointer\nFire turret: Left click\nCharge Jumpdrive: Hold <Space>",
                        new Vector2(GraphicsDevice.Viewport.Width / 2 - 250, 175 - verticalOffset),
                        Color.White);

                    spriteBatch.DrawString(
                        fontStatus,
                        "Hit asteroids to repel them-\nShields and turret will slowly recharge-\nCharging your jump drive takes 10 seconds-\nMeanwhile your shields and turret will\ngo offline leaving you unprotected-\n\nGood luck!",
                        new Vector2(GraphicsDevice.Viewport.Width / 2 - 250, 475 - verticalOffset),
                        Color.White);
                    break;
                case GameState.InGame:
                    spriteBatch.DrawString(
                        fontStatus,
                        String.Format(
                            "Shields: F{0:F0} R{1:F0} B{2:F0} L{3:F0}",
                            playerShip.Shields[0].Energy,
                            playerShip.Shields[1].Energy,
                            playerShip.Shields[2].Energy,
                            playerShip.Shields[3].Energy
                            ),
                        new Vector2(5, 5 - verticalOffset),
                        Color.White);

                    spriteBatch.DrawString(
                        fontStatus,
                        String.Format(
                            "Turret: {0:F0}",
                            playerShip.Turret.Energy
                            ),
                        new Vector2(5, 25 - verticalOffset),
                        Color.White);

                    spriteBatch.DrawString(
                        fontStatus,
                        String.Format(
                            "Drive: {0:F0}",
                            playerShip.DriveEnergy
                            ),
                        new Vector2(5, 45 - verticalOffset),
                        Color.White);

                    if (playerShip.ChargingDrive 
                        && !playerShip.IsJumped
                        && playerShip.IsAlive)
                    {
                        DrawStillCharging(gameTime);
                    }

                    if (!playerShip.IsAlive)
                    {
                        DrawHullBreach(gameTime);
                    }
                    break;
                case GameState.PostGame:
                    spriteBatch.DrawString(
                    fontStatus,
                    String.Format(
                        "Game Over",
                        playerShip.Turret.Energy
                        ),
                    new Vector2(GraphicsDevice.Viewport.Width / 2 - 65, 125 - verticalOffset),
                    Color.White);

                    if (playerShip.IsAlive)
                    {
                        DrawJumpComplete(gameTime);

                        spriteBatch.DrawString(
                            fontStatus,
                            "Congratulations, I hope it wasn't too easy!\n\nI had a lot of fun creating my first LD48 entry\nand maybe you liked it, too?\n\nPlease leave me some feedback!\n\nxfxian\n\n(Press <R> to start again)",
                            new Vector2(GraphicsDevice.Viewport.Width / 2 - 250, 175 - verticalOffset),
                            Color.White);
                    }
                    else
                    {
                        DrawHullBreach(gameTime);

                        spriteBatch.DrawString(
                            fontStatus,
                            "Press <R> to try again",
                            new Vector2(GraphicsDevice.Viewport.Width / 2 - 250, 175 - verticalOffset),
                            Color.White);
                    }

                    break;
            }

#if DEBUG

            Matrix proj = Matrix.CreateOrthographicOffCenter(
                0f, 
                ConvertUnits.ToSimUnits(GraphicsDevice.Viewport.Width),
                ConvertUnits.ToSimUnits(GraphicsDevice.Viewport.Height), 
                0f, 
                0f, 
                1f
                );

            var debugCamera = Matrix.CreateTranslation(new Vector3(
                0.0f,
                ConvertUnits.ToSimUnits(verticalOffset),
                0.0f
                ))
                ;

            debugView.RenderDebugData(ref proj, ref debugCamera);
#endif

            spriteBatch.End();

            base.Draw(gameTime);
        }

        private void DrawHullBreach(GameTime gameTime)
        {
            spriteBatch.DrawString(
            fontStatus,
            "Hull Breach",
            new Vector2(GraphicsDevice.Viewport.Width / 2 - 75, 100 - verticalOffset),
            Color.Red * ((float)gameTime.TotalGameTime.TotalSeconds % 1));
        }

        private void DrawJumpComplete(GameTime gameTime)
        {
            spriteBatch.DrawString(
            fontStatus,
            "Jump Complete",
            new Vector2(GraphicsDevice.Viewport.Width / 2 - 90, 100 - verticalOffset),
            Color.Green * ((float)gameTime.TotalGameTime.TotalSeconds % 1));
        }

        private void DrawPressStart(GameTime gameTime)
        {
            spriteBatch.DrawString(
            fontStatus,
            "Press <R> to start",
            new Vector2(GraphicsDevice.Viewport.Width / 2 - 95, 100 - verticalOffset),
            Color.LightBlue * ((float)gameTime.TotalGameTime.TotalSeconds % 1));
        }

        private void DrawStillCharging(GameTime gameTime)
        {
            spriteBatch.DrawString(
            fontStatus,
            String.Format(
                "Still charging! T-{0:F1}",
                10.0f - (playerShip.DriveEnergy / 100)
                ),
            new Vector2(GraphicsDevice.Viewport.Width / 2 - 95, 100 - verticalOffset),
            Color.Yellow * ((float)gameTime.TotalGameTime.TotalSeconds % 1));
        }
    }
}
